#!/usr/bin/env python3
# encoding: utf-8
'''
@author: rsu
'''

import sys, os, logging, string, math, json, re
sys.path.append( "/opt/essentia-200404-tf/lib64/python3.6/site-packages")
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

from collections import namedtuple
from pathlib import Path, PurePath
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from time import time

import numpy as np
from essentia.standard import *
import essentia


CFG_MFILEPATTERN = r'.*.(?:mp3|flac|m4a|ogg|wav)'
CFG_JSON_PATH = '/tmp'
#CFG_MODELS_PATH = '/home/luser/tmp/ess-test/models'
CFG_MODELS_PATH = os.getcwd()
#CFG_PATH_REPLACE = '/home/luser/tmp/sshfs:/data/music'
CFG_PATH_REPLACE = None
PATT_MODEL = re.compile( r'^[a-z_]+-(?:musicnn|vggish)-(:?msd|mtt|audioset)\.pb$')
PATT_MODEL_CNN = re.compile( r'.*-musicnn-.*')
PATT_MODEL_VGG = re.compile( r'.*-vggish-.*')

# Our models take audio streams at 16kHz
CFG_SAMPLERATE = 16000
CFG_REPLAYGAIN = True
CFG_EQUALLOUDNESS = True

# 'msd-musicnn': ['rock','pop','alternative','indie','electronic','female vocalists','dance','00s','alternative rock','jazz','beautiful','metal','chillout','male vocalists','classic rock','soul','indie rock','Mellow','electronica','80s','folk','90s','chill','instrumental','punk','oldies','blues','hard rock','ambient','acoustic','experimental','female vocalist','guitar','Hip-Hop','70s','party','country','easy listening','sexy','catchy','funk','electro','heavy metal','Progressive rock','60s','rnb','indie pop','sad','House','happy'],

TFMODELS_LABELS = {
	'genre_dortmund': ["alternative", "blues", "electronic", "folkcountr", "funksoulrnb", "jazz", "pop", "raphiphop", "rock"],
	'genre_electronic': ["ambient", "dnb", "house", "techno", "trance"],
	'genre_rosamerica': ['classic', 'dance', 'hip hop', 'jazz','pop', 'rnb', 'rock', 'speech'],
	'genre_tzanetakis': ["blu", "cla", "cou", "dis", "hip", "jaz", "met", "pop", "reg", "roc"],
	'mood_acoustic': ["acoustic", "non_acoustic"],
	'mood_aggressive': ["aggressive", "not_aggressive"],
	'mood_electronic': ["electronic", "non_electronic"],
	'mood_happy': ["happy", "non_happy"],
	'mood_party': ["party", "non_party"],
	'mood_relaxed': ["relaxed", "non_relaxed"],
	'mood_sad': ["sad", "non_sad"],
	'danceability': ["danceable", "not_danceable"],
	'gender': ["female", "male"],
	'tonal_atonal': ["tonal", "atonal"],
	'voice_instrumental': ["instrumental", "voice"]
}

# functions
essfn_resample16k = Resample( outputSampleRate=CFG_SAMPLERATE)
essfn_replaygain = ReplayGain()
essfn_equalloudness = EqualLoudness()

AnalysisResults = namedtuple( 'AnalysisResults', [ 'modelname', 'classes_probabilities', 'proctime'])

__all__ = []
__version__ = '0.0.1'
__date__ = '2020-06-22'
__updated__ = '2020-06-28'

DEBUG = 1
TESTRUN = 0
PROFILE = 0

_LOGGER = logging.getLogger(__name__)


class Track( object):
	def __init__(self, path, analysis):
		self.path = path
		self.metadata = self.load_metadata( path)
		self.analysis = analysis

	def load_metadata( self, path):
		try:
			meta = MetadataReader(filename=path, failOnError=True)()
		except RuntimeError as e:
			_LOGGER.error("failed to load metadata of [{}]: {}".format(  path, repr(e)))
			return None
		pool_meta, duration, bitrate, samplerate, channels = meta[7:]
		tags = {}
		for key in pool_meta.descriptorNames():
			tags[ key.replace('metadata.tags.', '')] = ",".join( pool_meta[key])
		return {
			"duration": duration,
			"bitrate": bitrate,
			"samplerate": samplerate,
			"channels": channels,
			"tags": tags
		}

	def write_json( self, jsondir, json_key_prefix, pathreplace=None):
		if pathreplace:
			if pathreplace.count( ':') != 1:
				_LOGGER.error( "gen_json: invalid replacement: {} ".format( pathreplace))
				return False
			(oldpath, newpath) = pathreplace.split( ':')
		highlevel = {}
		lowlevel = {}
		metadata = {}
		metadata[ 'audio_properties'] = { 'length': self.metadata[ 'duration']}
		metadata[ 'tags'] = { 'file_path': self.path if not pathreplace else self.path.replace( oldpath, newpath)}
		metadata[ 'tags']['file_name'] = os.path.basename( self.path)
		for tag in [ 'album', 'artist', 'genre', 'title']:
			metadata[ 'tags'][ tag] = [ self.metadata[ 'tags'][ tag]] if tag in self.metadata[ 'tags'] else ["Unknown"]
		# classes[ np.argmax( predAvg)]] + predAvg.tolist()
		for ar in self.analysis:
			json_key = json_key_prefix+ar.modelname
			highlevel[ json_key] = {}
			highlevel[ json_key]['probability'] = float( ar.classes_probabilities[0][1])
			highlevel[ json_key]['value'] = ar.classes_probabilities[0][0]
			highlevel[ json_key]['all'] = {}
			for cp in ar.classes_probabilities:
				highlevel[ json_key]['all'][ cp[0]] = float( cp[1])
	
		res = { 'highlevel': highlevel, 'metadata': metadata, 'lowlevel': lowlevel}
		_LOGGER.debug( "gen_json: {} ".format( repr( res)))
		jpath = os.path.join( jsondir, str( self.path).replace("/", "_") + ".json")
		_LOGGER.debug( "gen_json: file {} ".format( jpath))
		with open( jpath, mode="w") as data_file:
			json.dump( res, data_file)
		return res


def run_model( model_dir, mname, audio):
	res = []
	for submodel in gen_model_permutations( mname):
		model_file = os.path.join( model_dir, submodel+'.pb')
		if not os.path.isfile( model_file):
			#_LOGGER.debug("file {} not found".format( model_file))
			continue
		_LOGGER.debug("processing model: {} ".format( submodel))
		classes_probabilities = []
		start_time = time()
		try:
			if PATT_MODEL_CNN.match( submodel):
				predictions = TensorflowPredictMusiCNN(graphFilename=model_file)(audio)
			elif PATT_MODEL_VGG.match( submodel):
				predictions = TensorflowPredictVGGish(graphFilename=model_file)(audio)
			else:
				_LOGGER.error("unknown model type for: {}".format( submodel))
				return False
		except Exception as e:
			_LOGGER.error("fatal error in Tensorflow: {}".format( repr(e)))
			return None
		#_LOGGER.debug('model: {} processing time: {:.2f}s'.format( submodel, time() - start_time))
		if predictions is not None and predictions.any():
			# The shape of the predictions matrix is [n_patches, n_labels]
			# Take advantage of NumPy to average them over the time axis
			averaged_predictions = np.mean(predictions, axis=0)
			order = averaged_predictions.argsort()[::-1]
			for i in order:
				classes_probabilities.append( ( TFMODELS_LABELS[ mname][i], averaged_predictions[i]))
				#_LOGGER.debug('{}: {:.3f}'.format( TFMODELS_LABELS[ mname][i], averaged_predictions[i]))
			cidx = submodel.find('-')
			res.append( AnalysisResults( submodel[:cidx]+"."+submodel[cidx+1:], classes_probabilities, time() - start_time))
		else:
			_LOGGER.warn("no predictions for: {}".format( submodel))
	if res:
		_LOGGER.debug("result for model {}: {}".format( mname, repr( res)))
	return res

def gen_model_permutations( mname):
	return [ mname+'-vggish-audioset', mname+'-musicnn-msd', mname+'-musicnn-mtt']

def load_audio( filename):
	start_time = time()
	# Instantiate a MonoLoader and run it in the same line
	sampleRate = 44100
	try:
		audio = MonoLoader(filename=filename)()
		#audio = MonoLoader(filename=filename, sampleRate=CFG_SAMPLERATE)()
	except Exception as e:
		_LOGGER.error("fatal error reading audio: {}".format( repr(e)))
		return None, 0

	if CFG_REPLAYGAIN:
		# compute replay gain first
		replayGain = essfn_replaygain( audio) #(sampleRate = CFG_SAMPLERATE)(audio)
		# rescale audio if not silent (also apply a 6dB pre-amplification)
		if replayGain < 68.0:
			audio = Scale(factor = 10**((replayGain)/20))(audio)
	if CFG_EQUALLOUDNESS:
		audio = essfn_equalloudness( audio) #( sampleRate = CFG_SAMPLERATE)(audio)
	audio = essfn_resample16k ( audio)
	t = time() - start_time
	_LOGGER.debug("loaded audio file {} ({:.2f}s)".format( filename, t))
	return audio, t


def find_files( fpath, check_exist=False):
	if os.path.isfile( fpath):
		f = get_files_inputfile( fpath, check_exist=check_exist)
	elif os.path.isdir( fpath):
		f = get_files_dir( fpath)
	else:
		_LOGGER.error( "path {} not found".format( fpath))
		return None
	return f

def get_files_dir( fpath):
	fpatt = re.compile( CFG_MFILEPATTERN, re.IGNORECASE)
	allfiles = []

	for root, dirs, filenames in os.walk( fpath, followlinks=True):
		for f in filenames:
			if fpatt.match( f):
				allfiles.append( os.path.join(root, f))

	_LOGGER.debug( "get_files_dir: {} files from {}".format( len(allfiles), fpath))
	return allfiles

def get_files_inputfile( ifile, check_exist=False):
	allfiles = []
	fpatt = re.compile( '^#.*')

	with open( ifile, 'r') as fp:
		for line in fp:
			if not fpatt.match( line):
				if not check_exist or os.path.isfile( line.strip()):
					allfiles.append( line.strip())
		fp.close()

	_LOGGER.debug( "get_files_inputfile: {} files from {}".format( len(allfiles), ifile))
	return allfiles

class CLIError(Exception):
	'''Generic exception to raise and log different fatal errors.'''
	def __init__(self, msg):
		super(CLIError).__init__(type(self))
		self.msg = "E: %s" % msg
	def __str__(self):
		return self.msg
	def __unicode__(self):
		return self.msg

def main(argv=None): # IGNORE:C0111
	'''Command line options.'''

	if argv is None:
		argv = sys.argv
	else:
		sys.argv.extend(argv)

	program_name = os.path.basename(sys.argv[0])
	program_version = "v%s" % __version__
	program_build_date = str(__updated__)
	program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
	program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
	program_license = '''%s

	esstf-extractor v%s.
	Copyright 2020 R.S.U. GPL v3. All rights reserved.
	https://www.nexus0.net/pub/sw/lmsessentia/

USAGE
''' % (program_shortdesc, str(__version__))

	try:
		# Setup argument parser
		parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
		parser.add_argument('-V', '--version', action='version', version=program_version_message)
		parser.add_argument('--log-level', action='store', choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'], default='INFO', help="set log level (default: %(default)s)")

		parser.add_argument('source', action='store', help='file or directory of files to analyze')
		parser.add_argument('destdir', action='store', help='destination directory for analysis files')
		parser.add_argument('--model-dir', action='store', default=CFG_MODELS_PATH, help='models to apply (default: %(default)s)')
		parser.add_argument('--replace-pathname', action='store', default=CFG_PATH_REPLACE, help='replace pathname. Format: old_path:new_path')
		parser.add_argument('--no-skip-existing', action='store_false', help='do not skip audio file even if JSON output file exists')
		parser.add_argument('--json-key-prefix', action='store', default='esstf.', help='prefix for classes in JSON output(default: %(default)s)')
		#parser.add_argument('--no-normalize', action='store_false', help='do not normalize audio')
		#parser.add_argument('--no-equalloudness', action='store_false', help='do not equalize loudness')

		# Process arguments
		args = parser.parse_args()
		logging.basicConfig(level=args.log_level)
		
		if not os.path.isdir( args.destdir):
			_LOGGER.error("{} is not a directory".format( args.destdir))
			return -1

		all_models = [f.name for f in os.scandir( args.model_dir) if f.is_file() and PATT_MODEL.match( f.name)]
		_LOGGER.info("found {} models in {}".format ( len( all_models), args.model_dir))
		_LOGGER.debug("models found: {}".format ( "|".join( all_models)))
		if len( all_models) == 0:
			_LOGGER.error("No model found".format( args.destdir))
			return -1

		files_all = find_files( args.source)
		if not files_all or len( files_all) == 0:
			_LOGGER.warn("no files found")
			return -1
		stat_files_all = len( files_all)
		stat_files_ok = 0
		stat_files_skip = 0
		stat_files_err = 0

		_LOGGER.info("found {} files to process".format( stat_files_all))
		for fname in files_all:
			if not os.path.isfile( fname):
				_LOGGER.warn("file {} not found".format( fname))
				stat_files_err += 1
				continue
			if args.no_skip_existing:
				if os.path.isfile( os.path.join( args.destdir, str( fname).replace("/", "_") + ".json")):
					_LOGGER.info("skipping audio file {}: analyis result exists".format( fname))
					stat_files_skip += 1
					continue
			_LOGGER.debug("start processing audio file {}".format( fname))
			audio, al_time = load_audio( fname)
			if audio is None:
				stat_files_err += 1
				continue
			ar = []
			for mod in TFMODELS_LABELS:
				single_analysis = run_model( args.model_dir, mod, audio)
				if single_analysis is not None:
					ar += single_analysis
			if ar:
				t = Track( fname, ar)
				if t.metadata is not None:
					#_LOGGER.debug("meta: {}".format (repr( t.metadata['tags'])))
					t.write_json( args.destdir, args.json_key_prefix, pathreplace=args.replace_pathname)
					stat_files_ok += 1
					_LOGGER.info("processed file {}/{}: {} (time: audio: {:.2f}s / models ({}): {:.2f}s)".format( stat_files_ok+stat_files_skip+stat_files_err,  stat_files_all, fname, al_time, len( t.analysis), sum([ a.proctime for a in t.analysis])))
				else:
					stat_files_err += 1
			else:
				stat_files_err += 1
		_LOGGER.info("finished: {} files processed (ok: {} skipped: {} errors: {})".format( stat_files_all, stat_files_ok, stat_files_skip, stat_files_err))

	except KeyboardInterrupt:
		### handle keyboard interrupt ###
		return 0
	except Exception as e:
		if DEBUG or TESTRUN:
			raise(e)
		indent = len(program_name) * " "
		sys.stderr.write(program_name + ": " + repr(e) + "\n")
		sys.stderr.write(indent + "  for help use --help")
		return 2

if __name__ == "__main__":
	sys.exit(main())
